export const x00 = "x00";
export const x01 = "x01";
export const x02 = "x02";
export const x03 = "x03";
export const x04 = "x04";
export const x05 = "x05";
export const x06 = "x06";
export const x07 = "x07";
export const x08 = "x08";
export const x09 = "x09";
export const x10 = "x10";
export const x11 = "x11";
export const x12 = "x12";
export const x13 = "x13";
export const x14 = "x14";
export const x15 = "x15";
export const x16 = "x16";
export const x17 = "x17";
export const x18 = "x18";
export const x19 = "x19";

export const y00 = "y00";
export const y01 = "y01";
export const y02 = "y02";
export const y03 = "y03";
export const y04 = "y04";
export const y05 = "y05";
export const y06 = "y06";
export const y07 = "y07";
export const y08 = "y08";
export const y09 = "y09";
export const y10 = "y10";
export const y11 = "y11";
export const y12 = "y12";
export const y13 = "y13";
export const y14 = "y14";
export const y15 = "y15";
export const y16 = "y16";
export const y17 = "y17";
export const y18 = "y18";
export const y19 = "y19";
