export const abc = "abc";
export const def = "def";
export let toString;
export const setToString = () => {
	toString = () => "toString";
};
export const moduleId = module.id;
export const a = "single char";
export const $1 = "double char";
export const __1 = "3 chars";
