import value from "./dep-shared3";
import value4 from "./dep-shared4";

export function setOther2(value) {
	other2 = value;
}

export const value2 = value;
export const value3 = value4;
export var other2;

if (other2 === undefined) other2 = "other";
