module.exports = [
	[
		/export 'named' \(imported as 'named'\) was not found in '\.\/module' \(possible exports: default\)/
	],
	[
		/export 'named' \(imported as 'named2'\) was not found in '\.\/esModule' \(possible exports: __esModule, default\)/
	]
];
