module.exports = [
	{ code: /DEP_WEBPACK_DEPRECATION_ARRAY_TO_SET/ },
	{ code: /DEP_WEBPACK_DEPRECATION_ARRAY_TO_SET_INDEXER/ },
	{ code: /DEP_WEBPACK_DEPRECATION_ARRAY_TO_SET_LENGTH/ },
	{ code: /DEP_WEBPACK_DEPRECATION_ARRAY_TO_SET_PUSH/ }
];
