export { default as a } from "./a";
export { default as b } from "./b";
export { default as empty } from "./empty";
export { default as json } from "./json";

throw new Error("Should not be loaded");
