exports.module1 = require("./module?we1" + __resourceQuery);
var m2 = require("./module?we2" + __resourceQuery);
exports.module2 = m2;
this.module3 = require("./module?we3" + __resourceQuery);
var m4 = require("./module?we4" + __resourceQuery);
this.module4 = m4;
