/** @type {import("../../../../").Configuration} */
module.exports = {
	output: {
		libraryTarget: "system"
	},
	externals: {
		external1: "external1",
		external2: "external2",
		external3: "external3",
		external4: "external4",
		external5: "external5"
	}
};
