Object.defineProperty(exports, "__esModule", { get: () => true });
Object.defineProperty(exports, "default", { get: () => "default" });
Object.defineProperty(exports, "named", { get: () => "named" });
