module.exports = [
	[
		/conflicting star exports/,
		/names 'named1', 'named2', 'named3' /,
		/\.\/a\.js/,
		/\.\/b\.js/
	],
	[
		/conflicting star exports/,
		/names 'named1', 'named2', 'named3' /,
		/\.\/a\.js/,
		/\.\/b\.js/
	]
];
