module.exports = [
	[/export 'global' \(imported as 'style'\) was not found/],
	[/export 'nested2' \(imported as 'style'\) was not found/],
	[/export 'global-color' \(imported as 'style'\) was not found/],
	[/export 'global' \(imported as 'style'\) was not found/],
	[/export 'nested2' \(imported as 'style'\) was not found/],
	[/export 'global-color' \(imported as 'style'\) was not found/]
];
