module.exports = [
	{ code: /DEP_WEBPACK_CHUNK_ADD_MODULE/ },
	{ code: /DEP_WEBPACK_CHUNK_CONTAINS_MODULE/ },
	{ code: /DEP_WEBPACK_CHUNK_ENTRY_MODULE/ },
	{ code: /DEP_WEBPACK_CHUNK_GET_CHUNK_MODULE_MAPS/ },
	{ code: /DEP_WEBPACK_CHUNK_GET_MODULES/ },
	{ code: /DEP_WEBPACK_CHUNK_GET_NUMBER_OF_MODULES/ },
	{ code: /DEP_WEBPACK_CHUNK_HAS_ENTRY_MODULE/ },
	{ code: /DEP_WEBPACK_CHUNK_HAS_MODULE_IN_GRAPH/ },
	{ code: /DEP_WEBPACK_CHUNK_IS_EMPTY/ },
	{ code: /DEP_WEBPACK_CHUNK_MODULES_ITERABLE/ },
	{ code: /DEP_WEBPACK_CHUNK_MODULES_SIZE/ },
	{ code: /DEP_WEBPACK_CHUNK_REMOVE_MODULE/ },
	{ code: /DEP_WEBPACK_CHUNK_SIZE/ },
	{ code: /DEP_WEBPACK_MODULE_ADD_CHUNK/ },
	{ code: /DEP_WEBPACK_MODULE_CHUNKS_ITERABLE/ },
	{ code: /DEP_WEBPACK_MODULE_DEPTH/ },
	{ code: /DEP_WEBPACK_MODULE_GET_CHUNKS/ },
	{ code: /DEP_WEBPACK_MODULE_GET_NUMBER_OF_CHUNKS/ },
	{ code: /DEP_WEBPACK_MODULE_HASH/ },
	{ code: /DEP_WEBPACK_MODULE_INDEX2/ },
	{ code: /DEP_WEBPACK_MODULE_INDEX/ },
	{ code: /DEP_WEBPACK_MODULE_ISSUER/ },
	{ code: /DEP_WEBPACK_MODULE_IS_ENTRY_MODULE/ },
	{ code: /DEP_WEBPACK_MODULE_IS_IN_CHUNK/ },
	{ code: /DEP_WEBPACK_MODULE_OPTIMIZATION_BAILOUT/ },
	{ code: /DEP_WEBPACK_MODULE_OPTIONAL/ },
	{ code: /DEP_WEBPACK_MODULE_PROFILE/ },
	{ code: /DEP_WEBPACK_MODULE_REMOVE_CHUNK/ },
	{ code: /DEP_WEBPACK_MODULE_RENDERED_HASH/ },
	{ code: /DEP_WEBPACK_MODULE_SOURCE/ },
	{ code: /DEP_WEBPACK_MODULE_USED_EXPORTS/ }
];
