const createTestCases = require("../_helpers/createTestCases");
module.exports = createTestCases({
	nothing: {
		usedExports: [],
		expect: {
			"resize-observer": ["ResizeObserver"],
			"Source_Data/employee": [],
			"Source_Data/services": [],
			"Source_Framework/splitview/page": [],
			"Source_Framework/splitview/menu": [],
			"Source_Framework/splitview/text": [],
			"Source_Framework/uilistpicker/UiListPicker": [],
			"Source_ODS/cookiestorage": [],
			"Source_Framework/uirichtextedit": [],
			"Source_Framework/uielement": [],
			"Source_ODS/UiButton/icon": [],
			"Source_ODS/UiButton/text": [],
			"Source_Framework/pagelinks": [],
			"Source_ODS/ods": [],
			"Source_ODS/UiButton": [],
			"Source_Framework/defaultButtons": [],
			"Source_Framework/uiform": [],
			"Source_Framework/breadcrumb": [],
			"Source_ODS/lodash": [],
			"Source_ODS/DomElement": [],
			"Source_ODS/url": [],
			"Source_ODS/tooltip": [],
			"Source_ODS/ui": [],
			"Source_ODS/types": [],
			"Source_ODS/UiDialog": [],
			"Source_Framework/common/moduleDispatcher": [],
			"Source_Framework/common/employeeassignment": [],
			"Source_Framework/common/pdf_preview": [],
			"Source_Framework/common/taskprogress": [],
			"Source_Framework/uitable": [],
			"Source_Framework/backend/universalData": [],
			"Source_Framework/common/columnsDialog": [],
			"Source_Framework/backend/jsonData": [],
			"Source_Framework/common/statuschange": [],
			"Source_Framework/common/las_mod_jobselection": [],
			"./text": [],
			"./bec": [],
			"./positionTypes": [],
			"./settings": ["SETTINGS"],
			"./preferences": [],
			"./privileges": [],
			"./buttons": [],
			"./form": [],
			"./formHelper": [],
			"./table": [],
			"./techText": [],
			"./addPositionToSi": [],
			"./ciPosFromJob": [],
			"./shiftPosition": [],
			"./autoSums": [],
			"./defineLayout": [],
			"./insertFromTemplate": [],
			"./movePosition": [],
			"./copy": [],
			"./save": [],
			"./createOrder": [],
			"./orderPosFromJobPos": [],
			"./nchInfoDialog": [],
			"Source_Modules/las_mod_siheader/url": [],
			"Source_Modules/las_mod_siheader/bec": [],
			"Source_Modules/las_mod_ceheader/bec": [],
			"Source_Modules/las_mod_order/bec": [],
			"Source_Modules/las_mod_ciheader/bec": []
		}
	}
});
