module.exports = {
	target: "web",
	output: {
		chunkFilename: "[name].web.js",
		crossOriginLoading: "anonymous",
		trustedTypes: "customPolicyName"
	},
	performance: {
		hints: false
	},
	optimization: {
		minimize: false
	}
};
