/** @type {import("../../../../").Configuration} */
module.exports = {
	module: {
		rules: [
			{
				test: undefined,
				loader: undefined,
				use: undefined,
				options: undefined,
				resource: undefined
			}
		]
	}
};
