import {c} from "./c.js";

const b2 = 3;
const b3 = c;
export {c as b1, c as b4, b2, b3}
export const usedB1 = __webpack_exports_info__.b1.used;
export const usedB2 = __webpack_exports_info__.b2.used;
export const usedB3 = __webpack_exports_info__.b3.used;
export const usedB4 = __webpack_exports_info__.b4.used;
